/*
  MIT License
  Copyright (c) 2020 Michael Quinn

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

/*
  References
  Title: Z-Stack ZNP Interface Specification
  Author: Texas Instruments Incorporated
  Date: 2010-2012
  Revision: 1.4
  Availability: https://www.ti.com/tool/Z-STACK-ARCHIVE

  Title: Z-Stack Monitor and Test API
  Author: Texas Instruments Incorporated
  Date: 2008-2015
  Revision: 1.13
  Availability: https://www.ti.com/tool/Z-STACK-ARCHIVE

	Other
	Credit for the RECEIVE_SERIAL function goes to Robin2 from the Arduino forum.  The original code has been edited.
	Availability: https://forum.arduino.cc/t/serial-input-basics-updated/382007
*/

/*
  VT1300UART.h - Library
  Created by Michael Quinn

  Revision    Date          Descripton
  1.0.4       22/04/2021    Add device state
*/

#ifndef VT1300_Shield_h
	#define VT1300_Shield_h

  #if ARDUINO >= 100
    #include "Arduino.h"
    #else
    #include "WProgram.h"
  #endif

  /*
    Debug Mode
    Turn debugging off in Library to save program space.
  */
  #define DEBUG true //set to true for debug output, false for no debug output
  #define DEBUG_SERIAL if(DEBUG)Serial

  /*
    Class
    CC2530
  */
	class CC2530
	{
		public:

    /*
      Global Variables
    */
    const uint8_t NumBytes = 64;
    uint8_t ReceivedBytes[64];
    boolean NewData = false;
    boolean AFDataIncoming = false; // New AF_DATA_INOMING message
		uint8_t DeviceState = 0x00;
    int ZstackVersion = 0;

    CC2530(uint8_t PIN_CTS = 9, uint8_t PIN_RTS = 8, uint8_t PIN_RES = 7);
    void SKIPBOOT();
    void POWER_UP();
    void ZStackVersion();
    void COMMISSION();
    void WRITE_DATA(uint8_t *Data);
		void HARD_RESET_REQ();
    void SYS_RESET_REQ();
    void RECEIVE_SERIAL();
    void RECV_CALLBACK(unsigned long WaitTime);
    void RECV_TIMEOUT(unsigned long WaitTime);
    void EMPTY_BUFFER();
		void SRSP();
    boolean NEW_DATA();
    boolean AF_INCOMING_MSG();
    void LINK_QUALITY();
		void SYS_GPIO_SET_DIR(uint8_t Val);
		void SYS_GPIO_SET_INPUT_MODE(uint8_t Val);
		void SYS_GPIO_SET(uint8_t Val);
		void SYS_GPIO_CLEAR(uint8_t Val);
		void SYS_GPIO_READ(uint8_t Val);
		void AF_REGISTER(uint8_t EndPoint);
		void ZDO_STARTUP_FROM_APP(unsigned long LoopTime);
    void GET_SHORT_ADDRESS(uint8_t ShortAddr[2]);
    void GET_IEEE_ADDRESS(uint8_t IEEEAddr[8]);
    void ZDO_MGMT_PERMIT_JOIN_REQ(uint8_t Val = 0);
    void ZDO_END_DEVICE_BIND_REQ(uint8_t EndPoint);
    void ZDO_MGMT_LEAVE_REQ(uint8_t DstAddr[2], uint8_t IEEEAddr[8]);
    void ZDO_NODE_DESC_REQ(uint8_t DstAddr[2], uint8_t NWKAddrOfInterest[2]);
    uint16_t cmd_conv(uint8_t Cmd0, uint8_t Cmd1);
    uint8_t CALC_FCS(uint8_t *data, uint8_t len);
    void SetPANID(uint16_t Val = 0);
    void SetLOGICAL_TYPE(uint8_t Val = 0);
    void SetCHANLIST(uint8_t Val = 0);
    void SetZDO_DIRECT_CB(uint8_t Val = 0);
    void SetPOLL_RATE(uint32_t Val = 0);
    void SetQUEUED_POLL_RATE(uint16_t Val = 0);
    void SetRESPONSE_POLL_RATE(uint16_t Val = 0);
    void SetREJOIN_POLL_RATE(uint16_t Val = 0);
    void SetPOLL_FAILURE_RETRIES(uint8_t Val = 0);
    void SetPRECFGKEYEnable(uint8_t Val = 0);
    void SetPRECFGKEY(uint8_t a = 0, uint8_t b = 0, uint8_t c = 0, uint8_t d = 0, uint8_t e = 0, uint8_t f = 0, uint8_t g = 0, uint8_t h = 0, uint8_t i = 0, uint8_t j = 0, uint8_t k = 0, uint8_t l = 0, uint8_t m = 0, uint8_t n = 0, uint8_t o = 0, uint8_t p = 0);
    void SetAF_DATA_REQUEST(uint8_t DesEP = 0, uint8_t SourceEP = 0, uint8_t ClusterID0 = 0, uint8_t ClusterID1 = 0, uint8_t TransID = 0, uint8_t Options = 0, uint8_t Radius = 0);
    void SetAF_DATA_REQUEST_EXT(uint8_t DesEP = 0, uint8_t PanID0 = 0, uint8_t PanID1 = 0, uint8_t SourceEP = 0, uint8_t ClusterID0 = 0, uint8_t ClusterID1 = 0, uint8_t TransID = 0, uint8_t Options = 0, uint8_t Radius = 0);
    void SetTX_POWER(uint8_t Val = 0);
    void SetAPP_CNF_BDB_SET_TC_REQUIRE_KEY_EXCHANGE(uint8_t Val);

    /*
      AF_DATA_REQUEST
      Used by the Application processor to send a message
    */
    template <typename T> unsigned int AF_DATA_REQUEST (uint8_t ShortAddr0, uint8_t ShortAddr1, const T& Value, uint8_t Length)
    {
      uint8_t Sof = 0xFE;
      uint8_t Len = Length + 10;
      uint8_t Cmd0 = 0x24;
      uint8_t Cmd1 = 0x01;
      uint8_t DstAddr0 = ShortAddr1;
      uint8_t DstAddr1 = ShortAddr0;
      uint8_t DesEP = _AFDataReqCfg[0];
      uint8_t SourceEP = _AFDataReqCfg[1];
      uint8_t ClusterID0 = _AFDataReqCfg[2];
      uint8_t ClusterID1 = _AFDataReqCfg[3];
      uint8_t TransID = _AFDataReqCfg[4];
      uint8_t Options = _AFDataReqCfg[5];
      uint8_t Radius = _AFDataReqCfg[6];
      uint8_t DataLen = Length;

      // Make array
      uint8_t Data[14] = {Sof, Len, Cmd0, Cmd1, DstAddr0, DstAddr1, DesEP, SourceEP, ClusterID0, ClusterID1, TransID, Options, Radius, DataLen};
      uint8_t FCS0 = Len^Cmd0^Cmd1^DstAddr0^DstAddr1^DesEP^SourceEP^ClusterID0^ClusterID1^TransID^Options^Radius^DataLen;
      DEBUG_SERIAL.println(F("AF_DATA_REQUEST SREQ"));
      // SREQ
      digitalWrite(_RTS, HIGH);
      while (digitalRead(_CTS) == HIGH) {};

      Serial1.write(Data, sizeof(Data));

      const uint8_t * p = (const uint8_t*) &Value;
      unsigned int i;
      for (i = 0; i < Length; i++)
      Serial1.write(*p++);

      unsigned int FCS1 = 0;
      unsigned char *q = (unsigned char *)&Value;
      for (int i=0; i<Length; i++)
      {
        FCS1 ^= q[i];
      }

      uint8_t FinalFCS = FCS0^FCS1;
      Serial1.write(FinalFCS);

      Serial1.flush();
      digitalWrite(_RTS, LOW);

      return i;
    }


    /*
      AF_DATA_REQUEST_EXT
      Used for sending messages using binding.
      1. Send a ZDO_END_DEVICE_BIND_REQ(uint8_t EndPoint) on a End Device;
      2. Send a ZDO_END_DEVICE_BIND_REQ(uint8_t EndPoint) on the Coordinator within default 8 seconds Binding time.
      3. Send a AF_DATA_REQUEST_EXT (const T& Value) to lookup the address in the binding table
    */
    template <typename T> unsigned int AF_DATA_REQUEST_EXT (uint8_t AddrMode, const uint8_t IEEEAddr[8], const T& Value, uint8_t Length)
    {
      uint8_t Sof = 0xFE;
      uint8_t Len = Length + 20;
      uint8_t Cmd0 = 0x24;
      uint8_t Cmd1 = 0x02;
      uint8_t DstAddrMode = AddrMode; // Use destination mode 0x00 to lookup the address in the binding table
      uint8_t DstAddr0 = IEEEAddr[7];
      uint8_t DstAddr1 = IEEEAddr[6];
      uint8_t DstAddr2 = IEEEAddr[5];
      uint8_t DstAddr3 = IEEEAddr[4];
      uint8_t DstAddr4 = IEEEAddr[3];
      uint8_t DstAddr5 = IEEEAddr[2];
      uint8_t DstAddr6 = IEEEAddr[1];
      uint8_t DstAddr7 = IEEEAddr[0];
      uint8_t DesEP = _AFDataReqExtCfg[0];
      uint8_t DstPanId0 = _AFDataReqExtCfg[1];
      uint8_t DstPanId1 = _AFDataReqExtCfg[2];
      uint8_t SourceEP = _AFDataReqExtCfg[3];
      uint8_t ClusterID0 = _AFDataReqExtCfg[4];
      uint8_t ClusterID1 = _AFDataReqExtCfg[5];
      uint8_t TransID = _AFDataReqExtCfg[6];
      uint8_t Options = _AFDataReqExtCfg[7];
      uint8_t Radius = _AFDataReqExtCfg[8];
      uint8_t DataLen0 = Length;
      uint8_t DataLen1 = 0x00;

      // Make array
      uint8_t Data[24] = {Sof, Len, Cmd0, Cmd1, DstAddrMode, DstAddr0, DstAddr1, DstAddr2, DstAddr3, DstAddr4, DstAddr5, DstAddr6, DstAddr7, DesEP, DstPanId0, DstPanId1, SourceEP, ClusterID0, ClusterID1, TransID, Options, Radius, DataLen0, DataLen1};
      uint8_t FCS0 = Len^Cmd0^Cmd1^DstAddrMode^DstAddr0^DstAddr1^DstAddr2^DstAddr3^DstAddr4^DstAddr5^DstAddr6^DstAddr7^DesEP^DstPanId0^DstPanId1^SourceEP^ClusterID0^ClusterID1^TransID^Options^Radius^DataLen0^DataLen1;
      DEBUG_SERIAL.println(F("AF_DATA_REQUEST_EXT SREQ"));
      // SREQ
      digitalWrite(_RTS, HIGH);
      while (digitalRead(_CTS) == HIGH) {};

      Serial1.write(Data, sizeof(Data));

      const uint8_t * p = (const uint8_t*) &Value;
      unsigned int i;
      for (i = 0; i < Length; i++)
      Serial1.write(*p++);

      unsigned int FCS1 = 0;
      unsigned char *q = (unsigned char *)&Value;
      for (int i=0; i<Length; i++)
      {
        FCS1 ^= q[i];
      }

      uint8_t FinalFCS = FCS0^FCS1;
      Serial1.write(FinalFCS);

      Serial1.flush();
      digitalWrite(_RTS, LOW);

      return i;
    }


    /*
      COPY_PAYLOAD
      Copies the payload from the AF_INCOMING_MSG to a variable
    */
    template <typename T> unsigned int COPY_PAYLOAD (T& Value)
    {
      uint8_t *p;
      p = ReceivedBytes;
      memcpy(&Value, p + 20, ReceivedBytes[19]); // AFDataIncoming Data payload always starts at byte 20 in recieve buffer "ReceivedBytes"
    }

    private:

    uint8_t _CTS;
    uint8_t _RTS;
    uint8_t _RES;

    uint8_t _SYS_Reset[4] = {0x01, 0x41, 0x00, 0x00};
    uint8_t _TXPower[4] = {0x01, 0x21, 0x14, 0x04}; // Default 4dBm
    uint8_t _NVStartUpKeep[7] = {0x04, 0x21, 0x09, 0x03, 0x00, 0x00, 0x00}; // SYS_OSAL_NV_WRITE Keeps device specific and network parameters stored in non-volitile (NV) memory
    uint8_t _NVStartUpClear[8] = {0x05, 0x21, 0x09, 0x03, 0x00, 0x00, 0x01, 0x03}; // SYS_OSAL_NV_WRITE Clears device specific and network parameters stored in non-volitile (NV) memory
    uint8_t _PanID[9] = {0x06, 0x21, 0x09, 0x83, 0x00, 0x00, 0x02, 0xA1, 0x00}; // SYS_OSAL_NV_WRITE PanID Default 0x00A1
    uint8_t _Channel[11] = {0x08, 0x21, 0x09, 0x84, 0x00, 0x00, 0x04, 0x00, 0x08, 0x00, 0x00}; // SYS_OSAL_NV_WRITE Channel 11
    uint8_t _BDBSetPrimaryChannel[9] = {0x05, 0x2F, 0x08, 0x01, 0x00, 0x08, 0x00, 0x00}; // APP_CNF_BDB_SET_CHANNEL Default Primary Channel 11
    uint8_t _BDBSetSecondaryChannel[9] = {0x05, 0x2F, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00}; // APP_CNF_BDB_SET_CHANNEL Default Secondary Channel disabled
    uint8_t _BDBSetTCRequireKeyExchange[4] = {0x01, 0x2F, 0x09, 0x00}; // APP_CNF_BDB_SET_TC_REQUIRE_KEY_EXCHANGE Default False
    uint8_t _LogicalType[8] = {0x05, 0x21, 0x09, 0x87, 0x00, 0x00, 0x01, 0x00}; // SYS_OSAL_NV_WRITE Logical Type Default 0x00 Coordinator
    uint8_t _CallBack[8] = {0x05, 0x21, 0x09, 0x8F, 0x00, 0x00, 0x01, 0x01}; // SYS_OSAL_NV_WRITE Callback Default 0x01 True
    uint8_t _PollRate[11] = {0x08, 0x21, 0x09, 0x35, 0x00, 0x00, 0x04, 0xD0, 0x07, 0x00, 0x00}; // SYS_OSAL_NV_WRITE Poll rate Default 2000 milliseconds
    uint8_t _QueuedPollRate[9] = {0x06, 0x21, 0x09, 0x25, 0x00, 0x00, 0x02, 0x64, 0x00}; // SYS_OSAL_NV_WRITE Queued Poll Rate Default 100 milliseconds
    uint8_t _ResponsePollRate[9] = {0x06, 0x21, 0x09, 0x26, 0x00, 0x00, 0x02, 0x64, 0x00}; // SYS_OSAL_NV_WRITE Response Poll Rate Default 100 milliseconds
    uint8_t _RejoinPollRate[9] = {0x06, 0x21, 0x09, 0x27, 0x00, 0x00, 0x02, 0xB8, 0x01}; // SYS_OSAL_NV_WRITE Rejoin Poll Rate Default 440 milliseconds, required for End device join using TC Link Key, mandatory packets when joining network
    uint8_t _PollFailRetries[8] = {0x05, 0x21, 0x09, 0x29, 0x00, 0x00, 0x01, 0x03}; // SYS_OSAL_NV_WRITE Poll Failure Retries Default 0x03
    uint8_t _PreCFGKeyEnable[8] = {0x05, 0x21, 0x09, 0x63, 0x00, 0x00, 0x01, 0x01}; // SYS_OSAL_NV_WRITE Pre-configured Key enable Default 0x01 True
    uint8_t _PreCFGKey[23] = {0x14, 0x21, 0x09, 0x62, 0x00, 0x00, 0x10, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04}; // SYS_OSAL_NV_WRITE Pre-configured Key Default 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04
    uint8_t _AFDataReqCfg[7] = {0x01, 0x01, 0xB0, 0xFE, 0x01, 0x00, 0x04};
    uint8_t _AFDataReqExtCfg[9] = {0x01, _PanID[7], _PanID[8], 0x01, 0xB0, 0xFE, 0x01, 0x00, 0x04};
    uint8_t _ZDOStartUpFromApp[5] = {0x02, 0x25, 0x40, 0x00, 0x00};
    uint8_t _GPIOSetDir[6] = {0x02, 0x21, 0x0E, 0x00, 0x0F}; // Default 0x0F All GPIO's
    uint8_t _GPIOSetInput[6] = {0x02, 0x21, 0x0E, 0x01, 0x0F}; // Default 0x0F All GPIO's
    uint8_t _GPIOSet[6] = {0x02, 0x21, 0x0E, 0x02, 0x0F}; // Default 0x0F All GPIO's
    uint8_t _GPIOClear[6] = {0x02, 0x21, 0x0E, 0x03, 0x0F}; // Default 0x0F All GPIO's
    uint8_t _GPIORead[6] = {0x02, 0x21, 0x0E, 0x05, 0x0F}; // Default 0x0F All GPIO's
    uint8_t _NodeDesc[7] = {0x04, 0x25, 0x02, 0x00, 0x00, 0x00, 0x00};
    uint8_t _UtilGetDeviceInfo[3] = {0x00, 0x27, 0x00};
    uint8_t _PermitJoinOld[7] = {0x04, 0x25, 0x36, 0x00, 0x00, 0xFF, 0x01}; // Old version Z-stack 1.2.2
    uint8_t _PermitJoinNew[8] = {0x05, 0x25, 0x36, 0x02, 0x00, 0x00, 0xFF, 0x01}; // Updated in Z-stack 3.0.2
  };

#endif
